/* ***********************************************************************/
/*                                                                       */
/*                    WiNRADiO G65DDC EXCALIBUR SIGMA                    */
/*                         G65DDC device driver                          */
/*                      (c) Copyright 2023 RADIXON.                      */
/*                                                                       */
/* ***********************************************************************/

#include "g65ddc.h"


static struct class g65ddc_class=
{
    .name=G65DDC_MODULE_NAME
};

static dev_t base_device_number;
static struct mutex reg_mutex;
static struct g65ddc *registered_devices[G65DDC_MAX_DEVICE_COUNT];

static inline void reg_lock(void)
{
    mutex_lock(&reg_mutex);
}

static inline void reg_unlock(void)
{
    mutex_unlock(&reg_mutex);
}

static inline void dev_lock(struct g65ddc *dev)
{
    mutex_lock(&dev->mutex);
}

static inline void dev_unlock(struct g65ddc *dev)
{
    mutex_unlock(&dev->mutex);
}

#define SET_STATE(__dev__,__operation__) (__dev__)->state.s.__operation__;wake_up(&(__dev__)->state_wait_queue)

static inline void g65ddc_queue_wd_work(struct g65ddc *dev)
{
    queue_delayed_work(dev->workqueue,&dev->wd_delayed_work,HZ*2);
}

static inline void g65ddc_cancel_wd_work(struct g65ddc *dev)
{
    cancel_delayed_work_sync(&dev->wd_delayed_work);
}

static void g65ddc_wd(struct work_struct *work)
{
 struct g65ddc *dev;
 bool critical_temperature;
 int status;

    dev=container_of(work,struct g65ddc,wd_delayed_work.work);

    dev_lock(dev);

    if(dev->state.s.power && dev->state.s.connected && dev->open)
    {
        status=dev->fnc_check_temperature_locked(dev,&critical_temperature);

        for(;;)
        {
            if(likely(status==0))
            {
                if(unlikely(critical_temperature))
                {
                    dev->wd_counter++;

                    if(dev->wd_counter>=15)
                    {
                        dev_print_error(dev->p_dev,"critical temperature detected!");

                        dev->state.s.critical_temperature=1;
                        dev->fnc_set_power_locked(dev,0);

                        dev_print_warn(dev->p_dev,"device has been turned off!");
                        break;
                    }
                }
                else
                {
                    dev->wd_counter=0;
                }
            }
            else
            {
                dev->fnc_set_power_locked(dev,0);
                break;
            }

            g65ddc_queue_wd_work(dev);
            break;
        }
    }

    dev_unlock(dev);
}

static int g65ddc_ioctl_get_info(struct g65ddc *dev,void __user *user_ptr)
{
 int i;
 struct g65ddc_ioctl_info info;

    info.driver_version=INTERNAL_DRIVER_VERSION;
    info.bus_type=dev->bus_type;

    for(i=0;i<4;i++)
    {
        info.fw[i]=dev->fw[i];
        info.vid_pid[i]=dev->vid_pid[i];
    }

    if(unlikely(copy_to_user(user_ptr,&info,sizeof(info))))
    {
        return -EFAULT;
    }

    return 0;
}

static int g65ddc_ioctl_get_state(struct g65ddc *dev,struct file *file,void __user *user_ptr)
{
 union g65ddc_state state;
 int status;
 DEFINE_WAIT(wait);

    if(copy_from_user(&state,user_ptr,sizeof(state)))
    {
        return -EFAULT;
    }

    dev_lock(dev);

    for(;;)
    {
        if(state.value!=dev->state.value)
        {
            if(unlikely(copy_to_user(user_ptr,&dev->state,sizeof(state))))
            {
                status=-EFAULT;
            }
            else
            {
                status=0;
            }
            break;
        }

        if(!dev->open)
        {
            status=-EPERM;
            break;
        }

        if(file->f_flags & O_NONBLOCK)
        {
            status=-EAGAIN;
            break;
        }

        prepare_to_wait(&dev->state_wait_queue,&wait,TASK_INTERRUPTIBLE);

        dev_unlock(dev);

        schedule();

        dev_lock(dev);

        finish_wait(&dev->state_wait_queue,&wait);

        if(unlikely(signal_pending(current)))
        {
            status=-ERESTARTSYS;
            break;
        }
    }

    dev_unlock(dev);

    return status;
}

static int g65ddc_ioctl_set_power(struct g65ddc *dev,uint8_t power)
{
 int status;

    dev_lock(dev);

    if(power && dev->state.s.power==0)
    {
        dev->state.s.critical_temperature=0;
        dev->state.s.suspended=0;
        dev->wd_counter=0;
    }

    status=dev->fnc_set_power_locked(dev,power);

    if(power)
    {
        g65ddc_queue_wd_work(dev);
    }

    dev_unlock(dev);

    if(!power)
    {
        g65ddc_cancel_wd_work(dev);
    }

    return status;
}

static int g65ddc_ioctl(struct g65ddc *dev,struct file *file, unsigned int cmd,unsigned long arg)
{
    switch(cmd)
    {
        case G65DDC_IOCTL_GET_STATE:
            return g65ddc_ioctl_get_state(dev,file,(void __user*)arg);

        case G65DDC_IOCTL_SET_POWER:
            return g65ddc_ioctl_set_power(dev,(uint8_t)arg);

        case G65DDC_IOCTL_GET_INFO:
            return g65ddc_ioctl_get_info(dev,(void __user*)arg);

        default:
            return -ENOIOCTLCMD;
    }
}

#ifdef G65DDC_COMPAT

static int g65ddc_ioctl_compat(struct g65ddc *dev,struct file *file, unsigned int cmd,unsigned long arg)
{
    switch(cmd)
    {
        case G65DDC_IOCTL_GET_STATE:
            return g65ddc_ioctl_get_state(dev,file,compat_ptr(arg));

        case G65DDC_IOCTL_SET_POWER:
            return g65ddc_ioctl_set_power(dev,(uint8_t)arg);

        case G65DDC_IOCTL_GET_INFO:
            return g65ddc_ioctl_get_info(dev,compat_ptr(arg));

        default:
            return -ENOIOCTLCMD;
    }
}

static long g65ddc_fo_ioctl_compat(struct file *file, unsigned int cmd, unsigned long arg)
{
    if(unlikely(file->private_data==NULL))
    {
        return -ENODEV;
    }

    return ((struct g65ddc*)file->private_data)->fnc_ioctl_compat(file->private_data,file,cmd,arg);
}

#endif //G65DDC_COMPAT


static int g65ddc_fo_open(struct inode *node, struct file *file)
{
 uint32_t index;
 int status;
 struct g65ddc *dev;

    index=iminor(file->f_path.dentry->d_inode);

    if(index>=G65DDC_MAX_DEVICE_COUNT)
    {
        return -ENODEV;
    }

    reg_lock();

    dev=registered_devices[index];

    if(dev==NULL)
    {
        reg_unlock();
        return -ENODEV;
    }

    get_device(&dev->dev);

    dev_lock(dev);

    if(dev->open)
    {
        dev_unlock(dev);
        reg_unlock();
        put_device(&dev->dev);
        return -EBUSY;
    }

    dev->open=true;

    status=dev->fnc_open_locked(dev);

    if(status!=0)
    {
        dev->open=false;
        dev_unlock(dev);
        reg_unlock();
        put_device(&dev->dev);
        return status;
    }
    file->private_data=dev;

    dev_unlock(dev);

    reg_unlock();

    return 0;
}

static int g65ddc_fo_release(struct inode *node, struct file *file)
{
 struct g65ddc *dev;

    dev=file->private_data;

    if(dev!=NULL)
    {
        g65ddc_cancel_wd_work(dev);

        dev_lock(dev);

        dev->open=false;

        dev->fnc_close_locked(dev);

        wake_up_all(&dev->state_wait_queue);

        dev_unlock(dev);

        put_device(&dev->dev);

        file->private_data=NULL;
    }

    return 0;
}


static long g65ddc_fo_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    if(unlikely(file->private_data==NULL))
    {
        return -ENODEV;
    }

    return ((struct g65ddc*)file->private_data)->fnc_ioctl(file->private_data,file,cmd,arg);
}

static int g65ddc_fo_mmap(struct file *file,struct vm_area_struct *vma)
{
 struct g65ddc *dev;

    if(unlikely(file->private_data==NULL))
    {
        return -ENOTTY;
    }

    dev=file->private_data;

    if(unlikely(dev->fnc_mmap==NULL))
    {
        return -ENOTTY;
    }

    if(unlikely(!dev->state.s.connected))
    {
        return -ENODEV;
    }

    return dev->fnc_mmap(dev,vma);
}

static struct file_operations file_operations=
{
    .open=g65ddc_fo_open,
    .release=g65ddc_fo_release,
    .unlocked_ioctl=g65ddc_fo_ioctl,
    .mmap=g65ddc_fo_mmap,

#ifdef G65DDC_COMPAT
    .compat_ioctl=g65ddc_fo_ioctl_compat,
#endif
};

static int g65ddc_initialize(struct g65ddc *dev)
{
    dev->workqueue=create_singlethread_workqueue(G65DDC_MODULE_NAME);

    if(unlikely(dev->workqueue==NULL))
    {
        dev_print_error(dev->p_dev,"unable to create work queue!");
        return -ENOMEM;
    }

    INIT_DELAYED_WORK(&dev->wd_delayed_work,g65ddc_wd);

    mutex_init(&dev->mutex);
    init_waitqueue_head(&dev->state_wait_queue);

    dev->state.s.connected=1;

    return 0;
}


static void g65ddc_uninitialize(struct g65ddc *dev)
{
    if(dev->fnc_uninitialize)
    {
        dev->fnc_uninitialize(dev);
    }

    destroy_workqueue(dev->workqueue);

    mutex_destroy(&dev->mutex);
}

static void g65ddc_delete(struct g65ddc *dev)
{
    flush_workqueue(dev->workqueue);

    if(dev->fnc_delete)
    {
        dev->fnc_delete(dev);
    }

    g65ddc_uninitialize(dev);

    kfree(dev);
}

static void release_callback(struct device *dev)
{
 struct g65ddc *g65ddc_dev;

    print_dbg("release");

    g65ddc_dev=container_of(dev,struct g65ddc,dev);
    g65ddc_delete(g65ddc_dev);
}

static int g65ddc_register(struct g65ddc *dev)
{
 int status;
 int index;
 dev_t dev_no;

    reg_lock();

    for(index=0;index<G65DDC_MAX_DEVICE_COUNT;index++)
    {
        if(registered_devices[index]==NULL)
        {
            break;
        }
    }

    if(index>=G65DDC_MAX_DEVICE_COUNT)
    {
        reg_unlock();

        dev_print_error(dev->p_dev,"no free device slot available!\n");
        return -ENFILE;
    }

    dev->index=index;

    dev_no=MKDEV(MAJOR(base_device_number),index);

    cdev_init(&dev->cdev,&file_operations);

    dev->cdev.owner=THIS_MODULE;

    status=cdev_add(&dev->cdev,dev_no,1);

    if(status)
    {
        reg_unlock();
        dev_print_error(dev->p_dev,"char device registration failed!");
        return status;
    }

    memset(&dev->dev,0,sizeof(dev->dev));

    dev->dev.devt=dev_no;
    dev->dev.class=&g65ddc_class;
    dev->dev.release=release_callback;

    dev_set_name(&dev->dev,G65DDC_MODULE_NAME"%d",index);


    status=device_register(&dev->dev);

    if(status)
    {
        cdev_del(&dev->cdev);
        reg_unlock();
        dev_print_error(dev->p_dev,"failed to register device!");
        return status;
    }

    registered_devices[index]=dev;

    reg_unlock();

    return 0;
}


static void g65ddc_unregister(struct g65ddc *dev)
{
    reg_lock();

    if(registered_devices[dev->index]==dev)
    {
        registered_devices[dev->index]=NULL;

        cdev_del(&dev->cdev);
        device_unregister(&dev->dev);
    }

    reg_unlock();
}


static inline bool check_user_buffer(uintptr_t user_buffer,uint32_t size)
{
    return user_buffer!=0 && size!=0 && (user_buffer & (~PAGE_MASK))==0;
}


static int g65ddc_io_buffer_prepare(struct g65ddc_io_buffer *io_buffer,uintptr_t user_buffer,uint32_t size,bool write)
{
 int page_count,pinned,i;

    page_count=DIV_ROUND_UP(size,PAGE_SIZE);

    io_buffer->pages=kmalloc(page_count*sizeof(struct page*),GFP_KERNEL);

    if(unlikely(io_buffer->pages==NULL))
    {
        print_error("out of memory for io buffer pages!");
        return -ENOMEM;
    }

    io_buffer->sg_list=kmalloc(page_count*sizeof(struct scatterlist),GFP_KERNEL);

    if(unlikely(io_buffer->sg_list==NULL))
    {
        kfree(io_buffer->pages);
        print_error("out of memory for io buffer sg list!");
        return -ENOMEM;
    }

    pinned=get_user_pages_fast(user_buffer,page_count,write?1:0,io_buffer->pages);

    if(unlikely(pinned<0))
    {
        kfree(io_buffer->sg_list);
        kfree(io_buffer->pages);
        return -EFAULT;
    }

    if(unlikely(pinned<page_count))
    {
        for(i=0;i<pinned;i++)
        {
            put_page(io_buffer->pages[i]);
        }

        kfree(io_buffer->sg_list);
        kfree(io_buffer->pages);
        return -EFAULT;
    }

    io_buffer->user_buffer=user_buffer;
    io_buffer->size=size;
    io_buffer->page_count=pinned;
    io_buffer->status=0;
    io_buffer->idle=true;
    io_buffer->write=write;
    io_buffer->bytes_transferred=0;
    io_buffer->changed=false;

    sg_init_table(io_buffer->sg_list,page_count);

    for(i=0;i<page_count;i++)
    {
        if(size>=PAGE_SIZE)
        {
            sg_set_page(&io_buffer->sg_list[i],io_buffer->pages[i],PAGE_SIZE,0);
            size-=PAGE_SIZE;
        }
        else
        {
            sg_set_page(&io_buffer->sg_list[i],io_buffer->pages[i],size,0);
            break;
        }
    }

    return 0;
}

static void g65ddc_io_buffer_unprepare(struct g65ddc_io_buffer *io_buffer)
{
 int i;

    if(io_buffer->changed)
    {
        for(i=0;i<io_buffer->page_count;i++)
        {
            if(!PageReserved(io_buffer->pages[i]))
            {
                SetPageDirty(io_buffer->pages[i]);
            }
            put_page(io_buffer->pages[i]);
        }
    }
    else
    {
        for(i=0;i<io_buffer->page_count;i++)
        {
            put_page(io_buffer->pages[i]);
        }
    }

    kfree(io_buffer->sg_list);
    kfree(io_buffer->pages);
}



#ifdef G65DDC_USB

#define to_g65ddc_usb(__base_dev__) container_of(__base_dev__,struct g65ddc_usb,base_dev)

static inline void dev_usb_lock(struct g65ddc_usb *dev)
{
    dev_lock(&dev->base_dev);
}

static inline void dev_usb_unlock(struct g65ddc_usb *dev)
{
    dev_unlock(&dev->base_dev);
}

static inline void usb_lock_shared(struct g65ddc_usb *dev)
{
    down_read(&dev->rw_sem);
}

static inline void usb_unlock_shared(struct g65ddc_usb *dev)
{
    up_read(&dev->rw_sem);
}

static inline void usb_lock_exclusive(struct g65ddc_usb *dev)
{
    down_write(&dev->rw_sem);
}

static inline void usb_unlock_exclusive(struct g65ddc_usb *dev)
{
    up_write(&dev->rw_sem);
}


static int g65ddc_usb_ep_add_buffer(struct g65ddc_usb *dev,uint32_t ep_index,uintptr_t user_buffer,uint32_t size)
{
 struct g65ddc_usb_ep_buffer *buffer;
 unsigned long flags;
 int status;

    if(unlikely(!check_user_buffer(user_buffer,size) || ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    usb_lock_shared(dev);

    if(unlikely(dev->usb_interface==NULL))
    {
        usb_unlock_shared(dev);
        return -ENODEV;
    }

    buffer=kmalloc(sizeof(struct g65ddc_usb_ep_buffer),GFP_KERNEL);

    if(unlikely(buffer==NULL))
    {
        usb_unlock_shared(dev);
        dev_print_error(&dev->usb_dev->dev,"out of memory!");
        return -ENOMEM;
    }

    status=g65ddc_io_buffer_prepare(&buffer->io_buffer,user_buffer,size,usb_pipein(dev->ep[ep_index].pipe)!=0);

    if(unlikely(status!=0))
    {
        usb_unlock_shared(dev);
        kfree(buffer);
        return status;
    }

    buffer->urb=usb_alloc_urb(0,GFP_KERNEL);

    if(unlikely(buffer->urb==NULL))
    {
        usb_unlock_shared(dev);
        g65ddc_io_buffer_unprepare(&buffer->io_buffer);
        kfree(buffer);
        dev_print_error(&dev->usb_dev->dev,"out of memory!");
        return -ENOMEM;
    }

    buffer->ep=&dev->ep[ep_index];

    spin_lock_irqsave(&dev->ep[ep_index].spinlock,flags);

    list_add_tail(&buffer->io_buffer.list_entry,&dev->ep[ep_index].buffer_list);

    spin_unlock_irqrestore(&dev->ep[ep_index].spinlock,flags);

    usb_unlock_shared(dev);

    return 0;
}

static inline int g65ddc_usb_ioctl_ep_add_buffer(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_add_ep_buffer tmp;

    if(unlikely(copy_from_user(&tmp,user_ptr,sizeof(tmp))))
    {
        return -EFAULT;
    }

    return g65ddc_usb_ep_add_buffer(dev,tmp.ep_index,(uintptr_t)tmp.buffer,tmp.size);
}

static void g65ddc_usb_ep_cancel_buffers(struct g65ddc_usb *dev,struct g65ddc_usb_ep *ep)
{
    usb_lock_shared(dev);

    if(unlikely(dev->usb_interface!=NULL))
    {
        usb_kill_anchored_urbs(&ep->anchor);
    }

    usb_unlock_shared(dev);
}

static int g65ddc_usb_ioctl_ep_cancel_buffers(struct g65ddc_usb *dev,uint32_t ep_index)
{
    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    g65ddc_usb_ep_cancel_buffers(dev,&dev->ep[ep_index]);

    return 0;
}

static void g65ddc_usb_ep_release_buffer(struct g65ddc_usb_ep_buffer *buffer)
{
    g65ddc_io_buffer_unprepare(&buffer->io_buffer);
    usb_free_urb(buffer->urb);
    kfree(buffer);
}

static void g65ddc_usb_ep_release_buffers(struct g65ddc_usb_ep *ep)
{
 unsigned long flags;
 struct g65ddc_usb_ep_buffer *buffer;
 struct list_head buffer_list;

    INIT_LIST_HEAD(&buffer_list);

    spin_lock_irqsave(&ep->spinlock,flags);

    list_splice_init(&ep->buffer_list,&buffer_list);

    spin_unlock_irqrestore(&ep->spinlock,flags);

    list_for_each_entry(buffer,&buffer_list,io_buffer.list_entry)
    {
        g65ddc_usb_ep_release_buffer(buffer);
    }
}

static int g65ddc_usb_ioctl_ep_cancel_and_release_buffers(struct g65ddc_usb *dev,uint32_t ep_index)
{
 struct g65ddc_usb_ep *ep;

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    ep=dev->ep+ep_index;

    g65ddc_usb_ep_cancel_buffers(dev,ep);
    g65ddc_usb_ep_release_buffers(ep);

    return 0;
}

static void g65ddc_usb_urb_complete(struct urb *urb)
{
 struct g65ddc_usb_ep_buffer *buffer;
 struct g65ddc_usb_ep *ep;
 unsigned long flags;

    buffer=urb->context;
    ep=buffer->ep;

    spin_lock_irqsave(&ep->spinlock,flags);

    buffer->io_buffer.idle=true;
    buffer->io_buffer.status=urb->status;

    if(likely(buffer->io_buffer.status==0))
    {
        buffer->io_buffer.bytes_transferred=urb->actual_length;
    }

    if(list_first_entry_or_null(&ep->buffer_list,typeof(*buffer),io_buffer.list_entry)==buffer)
    {
        wake_up(&ep->wait_queue);
    }

    spin_unlock_irqrestore(&ep->spinlock,flags);
}

static int g65ddc_usb_ioctl_ep_bulk_requeue_buffer(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_ep_requeue_buffer requeue_buffer;
 struct g65ddc_usb_ep *ep;
 struct g65ddc_usb_ep_buffer *buffer;
 unsigned long flags;
 int status;

    if(unlikely(copy_from_user(&requeue_buffer,user_ptr,sizeof(requeue_buffer))))
    {
        return -EFAULT;
    }

    if(unlikely(requeue_buffer.ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    ep=&dev->ep[requeue_buffer.ep_index];

    if(unlikely(ep->type!=USB_ENDPOINT_XFER_BULK))
    {
        return -EINVAL;
    }

    if(unlikely(requeue_buffer.transfer_size<G65DDC_USB_EP_MIN_TRANSFER_SIZE))
    {
        requeue_buffer.transfer_size=G65DDC_USB_EP_MIN_TRANSFER_SIZE;
    }
    else if(unlikely(requeue_buffer.transfer_size>G65DDC_USB_EP_MAX_TRASNFER_SIZE))
    {
        requeue_buffer.transfer_size=G65DDC_USB_EP_MAX_TRASNFER_SIZE;
    }

    usb_lock_shared(dev);

    if(unlikely(dev->usb_interface==NULL))
    {
        usb_unlock_shared(dev);
        return -ENODEV;
    }

    spin_lock_irqsave(&ep->spinlock,flags);

    if(unlikely(list_empty(&ep->buffer_list)))
    {
        spin_unlock_irqrestore(&ep->spinlock,flags);
        usb_unlock_shared(dev);
        return 0;
    }

    buffer=list_first_entry(&ep->buffer_list,typeof(*buffer),io_buffer.list_entry);

    if(unlikely(!buffer->io_buffer.idle))
    {
        spin_unlock_irqrestore(&ep->spinlock,flags);
        usb_unlock_shared(dev);
        return 0;
    }

    list_move_tail(&buffer->io_buffer.list_entry,&ep->buffer_list);

    buffer->io_buffer.idle=false;
    buffer->io_buffer.bytes_transferred=0;
    buffer->io_buffer.status=0;

    requeue_buffer.transfer_size=min(buffer->io_buffer.size,requeue_buffer.transfer_size);

    usb_fill_bulk_urb(buffer->urb,dev->usb_dev,ep->pipe,NULL,requeue_buffer.transfer_size,
                              g65ddc_usb_urb_complete,buffer);

    buffer->urb->sg=buffer->io_buffer.sg_list;
    buffer->urb->num_sgs=buffer->io_buffer.page_count;

    usb_anchor_urb(buffer->urb,&ep->anchor);

    spin_unlock_irqrestore(&ep->spinlock,flags);

    status=usb_submit_urb(buffer->urb,GFP_KERNEL);

    if(unlikely(status))
    {
        buffer->io_buffer.idle=true;
        buffer->io_buffer.status=status;
        usb_unanchor_urb(buffer->urb);
    }

    usb_unlock_shared(dev);

    return status;
}

static inline int g65ddc_usb_set_led_pattern(struct g65ddc_usb *dev,uint16_t pattern)
{
 int status;

    usb_lock_shared(dev);

    if(likely(dev->usb_interface))
    {
        status=usb_control_msg(dev->usb_dev,dev->snd_ctrl_pipe,
                           0xC0,USB_TYPE_VENDOR|USB_DIR_OUT,pattern,0,NULL,0,G65DDC_USB_TIMEOUT);
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock_shared(dev);

    return status;
}

static int g65ddc_usb_update_led_pattern(struct g65ddc_usb *dev)
{
 uint16_t pattern;

    switch(dev->led_mode)
    {
        case USB_LED_MODE_OFF:
            pattern=USB_LED_PATTERN_OFF;
            break;

        case USB_LED_MODE_ON:
            pattern=USB_LED_PATTERN_ON;
            break;

        default:
            if(dev->base_dev.open)
            {
                pattern=dev->base_dev.state.s.power?USB_LED_PATTERN_POWER_ON:USB_LED_PATTERN_POWER_OFF;
            }
            else
            {
                pattern=USB_LED_PATTERN_APP_NOT_RUNNING;
            }
            break;
    }

    return g65ddc_usb_set_led_pattern(dev,pattern);
}

static inline int g65ddc_usb_clear_halt(struct g65ddc_usb *dev,struct g65ddc_usb_ep *ep)
{
 int status;

    usb_lock_shared(dev);

    if(likely(dev->usb_interface))
    {
        status=usb_clear_halt(dev->usb_dev,ep->pipe);
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock_shared(dev);

    return status;
}

static int g65ddc_usb_ioctl_clear_halt(struct g65ddc_usb *dev,uint32_t ep_index)
{
 int status;

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    usb_lock_shared(dev);

    if(likely(dev->usb_interface))
    {
        status=usb_clear_halt(dev->usb_dev,dev->ep[ep_index].pipe);
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock_shared(dev);

    return status;
}


static int g65ddc_usb_ioctl_reset_ep(struct g65ddc_usb *dev,uint32_t ep_index)
{
 int status;

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    usb_lock_shared(dev);

    if(likely(dev->usb_interface))
    {
        usb_reset_endpoint(dev->usb_dev,dev->ep[ep_index].address);

        status=0;
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock_shared(dev);

    return status;
}

static int g65ddc_usb_ctrl_vendor_msg_snd(struct g65ddc_usb *dev,uint8_t request,uint16_t value,uint16_t index,void __user *user_buffer,uint16_t size)
{
 void *buffer;
 int ret;

    if(unlikely((user_buffer!=NULL) ^ (size!=0)))
    {
        return -EINVAL;
    }

    if(user_buffer!=NULL)
    {
        buffer=kmalloc(size,GFP_KERNEL);

        if(unlikely(buffer==NULL))
        {
            return -ENOMEM;
        }

        if(unlikely(copy_from_user(buffer,user_buffer,size)))
        {
            kfree(buffer);
            return -EFAULT;
        }
    }
    else
    {
        buffer=NULL;
    }

    usb_lock_shared(dev);

    if(likely(dev->usb_interface))
    {
        ret=usb_control_msg(dev->usb_dev,dev->snd_ctrl_pipe,request,USB_TYPE_VENDOR|USB_DIR_OUT,value,index,buffer,size,G65DDC_USB_TIMEOUT);
    }
    else
    {
        ret=-ENODEV;
    }

    usb_unlock_shared(dev);

    if(buffer!=NULL)
    {
        kfree(buffer);
    }

    return ret;
}

static int g65ddc_usb_ctrl_vendor_msg_rcv(struct g65ddc_usb *dev,uint8_t request,uint16_t value,uint16_t index,void __user *user_buffer,uint16_t size)
{
 void *buffer;
 int ret;

    if(unlikely((user_buffer!=NULL) ^ (size!=0)))
    {
        return -EINVAL;
    }

    if(user_buffer!=NULL)
    {
        buffer=kmalloc(size,GFP_KERNEL);

        if(unlikely(buffer==NULL))
        {
            return -ENOMEM;
        }
    }
    else
    {
        buffer=NULL;
    }

    usb_lock_shared(dev);

    if(likely(dev->usb_interface))
    {
        ret=usb_control_msg(dev->usb_dev,dev->rcv_ctrl_pipe,request,USB_TYPE_VENDOR|USB_DIR_IN,value,index,buffer,size,G65DDC_USB_TIMEOUT);
    }
    else
    {
        ret=-ENODEV;
    }

    usb_unlock_shared(dev);

    if(buffer!=NULL)
    {
        if(ret>0)
        {
            if(unlikely(copy_to_user(user_buffer,buffer,ret)))
            {
                ret=-EFAULT;
            }
        }

        kfree(buffer);
    }

    return ret;
}

static inline int g65ddc_usb_ioctl_ctrl_vendor_msg_snd(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_ctrl_vendor_msg msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g65ddc_usb_ctrl_vendor_msg_snd(dev,msg.request,msg.value,msg.index,(void __user*)msg.data,msg.size);
}

static inline int g65ddc_usb_ioctl_ctrl_vendor_msg_rcv(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_ctrl_vendor_msg msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g65ddc_usb_ctrl_vendor_msg_rcv(dev,msg.request,msg.value,msg.index,(void __user*)msg.data,msg.size);
}

static int g65ddc_usb_bulk_msg_snd(struct g65ddc_usb *dev,uint32_t ep_index,void __user *user_buffer,uint32_t size)
{
 void *buffer;
 int transferred;
 int status;
 struct g65ddc_usb_ep *ep;

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    ep=dev->ep+ep_index;

    if(unlikely(user_buffer==NULL || size==0 || (ep->address&USB_DIR_IN) || ep->type!=USB_ENDPOINT_XFER_BULK))
    {
        return -EINVAL;
    }

    buffer=kmalloc(size,GFP_KERNEL);

    if(unlikely(buffer==NULL))
    {
        return -ENOMEM;
    }

    if(unlikely(copy_from_user(buffer,user_buffer,size)))
    {
        kfree(buffer);
        return -EFAULT;
    }

    usb_lock_shared(dev);

    if(likely(dev->usb_interface))
    {
        status=usb_bulk_msg(dev->usb_dev,ep->pipe,buffer,size,&transferred,G65DDC_USB_TIMEOUT);
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock_shared(dev);

    kfree(buffer);

    if(unlikely(status<0))
    {
        return status;
    }

    return transferred;
}


static int g65ddc_usb_bulk_msg_rcv(struct g65ddc_usb *dev,uint32_t ep_index,void __user *user_buffer,uint32_t size)
{
 void *buffer;
 int transferred;
 int status;
 struct g65ddc_usb_ep *ep;

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    ep=dev->ep+ep_index;

    if(unlikely(user_buffer==NULL || size==0 || (ep->address&USB_DIR_IN)==0 || ep->type!=USB_ENDPOINT_XFER_BULK))
    {
        return -EINVAL;
    }

    buffer=kmalloc(size,GFP_KERNEL);

    if(unlikely(buffer==NULL))
    {
        return -ENOMEM;
    }

    usb_lock_shared(dev);

    if(likely(dev->usb_interface))
    {
        status=usb_bulk_msg(dev->usb_dev,ep->pipe,buffer,size,&transferred,G65DDC_USB_TIMEOUT);
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock_shared(dev);

    if(unlikely(status<0))
    {
        kfree(buffer);
        return status;
    }

    if(unlikely(copy_to_user(user_buffer,buffer,transferred)))
    {
        kfree(buffer);
        return -EFAULT;
    }

    kfree(buffer);

    return transferred;
}

static inline int g65ddc_usb_ioctl_bulk_msg_rcv(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_bulk_msg msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g65ddc_usb_bulk_msg_rcv(dev,msg.ep_index,(void __user*)msg.data,msg.size);
}

static inline int g65ddc_usb_ioctl_bulk_msg_snd(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_bulk_msg msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g65ddc_usb_bulk_msg_snd(dev,msg.ep_index,(void __user*)msg.data,msg.size);
}

static int g65ddc_usb_ioctl_set_led_mode(struct g65ddc_usb *dev,uint32_t led_mode)
{
 int status;

    switch(led_mode)
    {
        case USB_LED_MODE_OFF:
        case USB_LED_MODE_ON:
        case USB_LED_MODE_DIAG:
            dev_usb_lock(dev);

            if(led_mode!=dev->led_mode)
            {
                dev->led_mode=led_mode;
                status=g65ddc_usb_update_led_pattern(dev);
            }
            else
            {
                status=0;
            }

            dev_usb_unlock(dev);

            return status;

        default:
            return -EINVAL;
    }

    return g65ddc_usb_update_led_pattern(dev);
}

static int g65ddc_usb_ioctl_get_ep_info(struct g65ddc_usb *dev,struct g65ddc_ioctl_usb_ep __user *user_ep)
{
 uint32_t i;
 struct g65ddc_ioctl_usb_ep ep;

    for(i=0;i<dev->ep_count;i++)
    {
        ep.address=dev->ep[i].address;
        ep.interval=dev->ep[i].interval;
        ep.max_packet_size=dev->ep[i].max_packet_size;
        ep.type=dev->ep[i].type;

        if(unlikely(copy_to_user(user_ep,&ep,sizeof(ep))))
        {
            return -EFAULT;
        }

        user_ep++;
    }

    return dev->ep_count;
}

static int g65ddc_usb_ep_get_complete_buffer(struct g65ddc_usb *dev,uint32_t ep_index,uintptr_t *user_buffer)
{
 struct g65ddc_usb_ep_buffer *buffer;
 struct g65ddc_usb_ep *ep;
 unsigned long flags;
 int status;
 DEFINE_WAIT(wait);

    if(unlikely(ep_index>=dev->ep_count))
    {
        return -EINVAL;
    }

    ep=&dev->ep[ep_index];

    for(;;)
    {
        usb_lock_shared(dev);

        if(unlikely(dev->usb_interface==NULL))
        {
            usb_unlock_shared(dev);
            return -ENODEV;
        }

        if(unlikely(dev->base_dev.state.s.power==0))
        {
            usb_unlock_shared(dev);
            return -EPERM;
        }

        spin_lock_irqsave(&ep->spinlock,flags);

        if(likely(!list_empty(&ep->buffer_list)))
        {
            buffer=list_first_entry(&ep->buffer_list,typeof(*buffer),io_buffer.list_entry);

            if(buffer->io_buffer.idle)
            {
                if(likely(buffer->io_buffer.status==0))
                {
                    *user_buffer=buffer->io_buffer.user_buffer;
                    status=buffer->io_buffer.bytes_transferred;
                }
                else
                {
                    status=buffer->io_buffer.status;
                }

                spin_unlock_irqrestore(&ep->spinlock,flags);
                usb_unlock_shared(dev);
                return status;
            }
        }

        prepare_to_wait(&ep->wait_queue,&wait,TASK_INTERRUPTIBLE);

        spin_unlock_irqrestore(&ep->spinlock,flags);

        usb_unlock_shared(dev);

        schedule();

        finish_wait(&ep->wait_queue,&wait);

        if(unlikely(signal_pending(current)))
        {
            return -ERESTARTSYS;
        }
    }
}

static inline int g65ddc_usb_ioctl_ep_get_complete_buffer(struct g65ddc_usb *dev,void __user *user_ptr)
{
 union g65ddc_ioctl_usb_ep_complete_buffer tmp;
 uintptr_t user_buffer;
 int status;

    if(unlikely(copy_from_user(&tmp.ep_index,user_ptr,sizeof(tmp.ep_index))))
    {
        return -EFAULT;
    }

    status=g65ddc_usb_ep_get_complete_buffer(dev,tmp.ep_index,&user_buffer);

    if(likely(status>=0))
    {
        if(unlikely(copy_to_user((void __user*)user_ptr,&user_buffer,sizeof(tmp.buffer))))
        {
            return -EFAULT;
        }
    }

    return status;
}

static long g65ddc_usb_ioctl(struct g65ddc *base_dev,struct file *file,unsigned int cmd,unsigned long arg)
{
    switch(cmd)
    {
        case G65DDC_IOCTL_USB_EP_GET_COMPLETE_BUFFER:
            return g65ddc_usb_ioctl_ep_get_complete_buffer(to_g65ddc_usb(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_USB_EP_REQUEUE_BUFFER:
            return g65ddc_usb_ioctl_ep_bulk_requeue_buffer(to_g65ddc_usb(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_USB_CTRL_VENDOR_MSG_RCV:
            return g65ddc_usb_ioctl_ctrl_vendor_msg_rcv(to_g65ddc_usb(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_USB_CTRL_VENDOR_MSG_SND:
            return g65ddc_usb_ioctl_ctrl_vendor_msg_snd(to_g65ddc_usb(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_USB_BULK_MSG_RCV:
            return g65ddc_usb_ioctl_bulk_msg_rcv(to_g65ddc_usb(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_USB_BULK_MSG_SND:
            return g65ddc_usb_ioctl_bulk_msg_snd(to_g65ddc_usb(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_USB_EP_ADD_BUFFER:
            return g65ddc_usb_ioctl_ep_add_buffer(to_g65ddc_usb(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_USB_EP_RELEASE_BUFFERS:
            return g65ddc_usb_ioctl_ep_cancel_and_release_buffers(to_g65ddc_usb(base_dev),(uint32_t)arg);

        case G65DDC_IOCTL_USB_EP_CANCEL_BUFFERS:
            return g65ddc_usb_ioctl_ep_cancel_buffers(to_g65ddc_usb(base_dev),(uint32_t)arg);

        case G65DDC_IOCTL_USB_EP_CLEAR_HALT:
            return g65ddc_usb_ioctl_clear_halt(to_g65ddc_usb(base_dev),(uint32_t)arg);

        case G65DDC_IOCTL_USB_EP_RESET:
            return g65ddc_usb_ioctl_reset_ep(to_g65ddc_usb(base_dev),(uint32_t)arg);

        case G65DDC_IOCTL_USB_SET_LED_MODE:
            return g65ddc_usb_ioctl_set_led_mode(to_g65ddc_usb(base_dev),(uint32_t)arg);

        case G65DDC_IOCTL_USB_GET_EP_INFO:
            return g65ddc_usb_ioctl_get_ep_info(to_g65ddc_usb(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_USB_GET_EP_COUNT:
            return to_g65ddc_usb(base_dev)->ep_count;

        default:
            return g65ddc_ioctl(base_dev,file,cmd,arg);
    }
}

#ifdef G65DDC_COMPAT

static inline int g65ddc_usb_ioctl_ep_get_complete_buffer_compat(struct g65ddc_usb *dev,void __user *user_ptr)
{
 union g65ddc_ioctl_usb_ep_complete_buffer_compat tmp;
 uintptr_t user_buffer;
 int status;

    if(unlikely(copy_from_user(&tmp.ep_index,user_ptr,sizeof(tmp.ep_index))))
    {
        return -EFAULT;
    }

    status=g65ddc_usb_ep_get_complete_buffer(dev,tmp.ep_index,&user_buffer);

    if(likely(status>=0))
    {
        if(unlikely(copy_to_user(user_ptr,&user_buffer,sizeof(tmp.buffer))))
        {
            return -EFAULT;
        }
    }

    return status;
}

static inline int g65ddc_usb_ioctl_ctrl_vendor_msg_rcv_compat(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_ctrl_vendor_msg_compat msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g65ddc_usb_ctrl_vendor_msg_rcv(dev,msg.request,msg.value,msg.index,compat_ptr(msg.data),msg.size);
}

static inline int g65ddc_usb_ioctl_ctrl_vendor_msg_snd_compat(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_ctrl_vendor_msg_compat msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g65ddc_usb_ctrl_vendor_msg_snd(dev,msg.request,msg.value,msg.index,compat_ptr(msg.data),msg.size);
}

static inline int g65ddc_usb_ioctl_bulk_msg_rcv_compat(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_bulk_msg_compat msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g65ddc_usb_bulk_msg_rcv(dev,msg.ep_index,compat_ptr(msg.data),msg.size);
}

static inline int g65ddc_usb_ioctl_bulk_msg_snd_compat(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_bulk_msg_compat msg;

    if(unlikely(copy_from_user(&msg,user_ptr,sizeof(msg))))
    {
        return -EFAULT;
    }

    return g65ddc_usb_bulk_msg_snd(dev,msg.ep_index,compat_ptr(msg.data),msg.size);
}

static inline int g65ddc_usb_ioctl_ep_add_buffer_compat(struct g65ddc_usb *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_usb_add_ep_buffer_compat tmp;

    if(unlikely(copy_from_user(&tmp,user_ptr,sizeof(tmp))))
    {
        return -EFAULT;
    }

    return g65ddc_usb_ep_add_buffer(dev,tmp.ep_index,(uintptr_t)compat_ptr(tmp.buffer),tmp.size);
}

static long g65ddc_usb_ioctl_compat(struct g65ddc *base_dev,struct file *file, unsigned int cmd, unsigned long arg)
{
    switch(cmd)
    {
        case G65DDC_IOCTL_USB_EP_GET_COMPLETE_BUFFER_COMPAT:
            return g65ddc_usb_ioctl_ep_get_complete_buffer_compat(to_g65ddc_usb(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_USB_EP_REQUEUE_BUFFER:
            return g65ddc_usb_ioctl_ep_bulk_requeue_buffer(to_g65ddc_usb(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_USB_CTRL_VENDOR_MSG_RCV_COMPAT:
            return g65ddc_usb_ioctl_ctrl_vendor_msg_rcv_compat(to_g65ddc_usb(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_USB_CTRL_VENDOR_MSG_SND_COMPAT:
            return g65ddc_usb_ioctl_ctrl_vendor_msg_snd_compat(to_g65ddc_usb(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_USB_BULK_MSG_RCV_COMPAT:
            return g65ddc_usb_ioctl_bulk_msg_rcv_compat(to_g65ddc_usb(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_USB_BULK_MSG_SND_COMPAT:
            return g65ddc_usb_ioctl_bulk_msg_snd_compat(to_g65ddc_usb(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_USB_EP_ADD_BUFFER_COMPAT:
            return g65ddc_usb_ioctl_ep_add_buffer_compat(to_g65ddc_usb(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_USB_EP_RELEASE_BUFFERS:
            return g65ddc_usb_ioctl_ep_cancel_and_release_buffers(to_g65ddc_usb(base_dev),(uint32_t)arg);

        case G65DDC_IOCTL_USB_EP_CANCEL_BUFFERS:
            return g65ddc_usb_ioctl_ep_cancel_buffers(to_g65ddc_usb(base_dev),(uint32_t)arg);

        case G65DDC_IOCTL_USB_EP_CLEAR_HALT:
            return g65ddc_usb_ioctl_clear_halt(to_g65ddc_usb(base_dev),(uint32_t)arg);

        case G65DDC_IOCTL_USB_EP_RESET:
            return g65ddc_usb_ioctl_reset_ep(to_g65ddc_usb(base_dev),(uint32_t)arg);

        case G65DDC_IOCTL_USB_SET_LED_MODE:
            return g65ddc_usb_ioctl_set_led_mode(to_g65ddc_usb(base_dev),(uint32_t)arg);

        case G65DDC_IOCTL_USB_GET_EP_INFO:
            return g65ddc_usb_ioctl_get_ep_info(to_g65ddc_usb(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_USB_GET_EP_COUNT:
            return to_g65ddc_usb(base_dev)->ep_count;

        default:
            return g65ddc_ioctl_compat(base_dev,file,cmd,arg);
    }
}

#endif //G65DDC_COMPAT

static int g65ddc_usb_set_power_locked(struct g65ddc *base_dev,uint8_t power)
{
 uint32_t i;
 uint8_t p;
 int status;
 struct g65ddc_usb *dev;

    if(base_dev->state.s.power==power)
    {
        return 0;
    }

    dev=to_g65ddc_usb(base_dev);

    usb_lock_exclusive(dev);

    if(unlikely(dev->usb_interface==NULL))
    {
        usb_unlock_exclusive(dev);
        return -ENODEV;
    }

    if(power==0)
    {
        p=base_dev->state.s.power;

        SET_STATE(base_dev,power=0);

        for(i=0;i<dev->ep_count;i++)
        {
            usb_kill_anchored_urbs(&dev->ep[i].anchor);
            wake_up(&dev->ep[i].wait_queue);
        }

        if((p & 0x01) && (p & 0x06))
        {
            usb_control_msg(dev->usb_dev,dev->snd_ctrl_pipe,
                       0xC7,USB_TYPE_VENDOR|USB_DIR_OUT,0x01,0,NULL,0,G65DDC_USB_TIMEOUT);
        }
    }

    status=usb_control_msg(dev->usb_dev,dev->snd_ctrl_pipe,
                       0xC7,USB_TYPE_VENDOR|USB_DIR_OUT,power,0,NULL,0,G65DDC_USB_TIMEOUT);

    usb_unlock_exclusive(dev);

    if(likely(status==0))
    {
        if(power)
        {
            SET_STATE(base_dev,power=power);
        }

        g65ddc_usb_update_led_pattern(dev);
    }

    return status;
}


static void g65ddc_usb_delete(struct g65ddc *base_dev)
{
 struct g65ddc_usb *dev;

    dev=to_g65ddc_usb(base_dev);

    usb_put_intf(dev->usb_interface);
    usb_put_dev(dev->usb_dev);
}

static int g65ddc_usb_open_locked(struct g65ddc *base_dev)
{
    g65ddc_usb_update_led_pattern(to_g65ddc_usb(base_dev));

    return 0;
}

static void g65ddc_usb_close_locked(struct g65ddc *base_dev)
{
 uint32_t i;
 struct g65ddc_usb *dev;

    g65ddc_usb_set_power_locked(base_dev,0);

    dev=to_g65ddc_usb(base_dev);

    usb_lock_shared(dev);

    for(i=0;i<dev->ep_count;i++)
    {
        g65ddc_usb_ep_release_buffers(&dev->ep[i]);
    }

    usb_unlock_shared(dev);
}

static int g65ddc_usb_check_temperature_locked(struct g65ddc *base_dev,bool *critical_temperature)
{
 int16_t *temps;
 int status;
 struct g65ddc_usb *dev;

    dev=to_g65ddc_usb(base_dev);

    temps=kmalloc(4*sizeof(int16_t),GFP_KERNEL);

    if(unlikely(temps==NULL))
    {
        dev_print_error(&dev->usb_dev->dev,"out of memory!");
        return -ENOMEM;
    }

    usb_lock_shared(dev);

    if(likely(dev->usb_interface!=NULL))
    {
        status=usb_control_msg(dev->usb_dev,dev->rcv_ctrl_pipe,0xCB,USB_TYPE_VENDOR|USB_DIR_IN,0,0,temps,4*sizeof(int16_t),G65DDC_USB_TIMEOUT);

        if(likely(status>=4*sizeof(int16_t)))
        {
            *critical_temperature=(int16_t)le16_to_cpu(temps[0])>=CRITICAL_TEMPERATURE_0 ||
                                  (int16_t)le16_to_cpu(temps[1])>=CRITICAL_TEMPERATURE_1 ||
                                  (int16_t)le16_to_cpu(temps[2])>=CRITICAL_TEMPERATURE_2;

            status=0;
        }
        else
        {
            status=-EIO;
        }
    }
    else
    {
        status=-ENODEV;
    }

    usb_unlock_shared(dev);

    kfree(temps);

    return status;
}

static void g65ddc_usb_uninitialize(struct g65ddc *base_dev)
{
}


static int g65ddc_usb_initialize(struct g65ddc_usb *dev)
{
 uint32_t i;
 int status;

    status=g65ddc_initialize(&dev->base_dev);

    if(unlikely(status!=0))
    {
        return status;
    }

    dev->base_dev.fnc_delete=g65ddc_usb_delete;
    dev->base_dev.fnc_uninitialize=g65ddc_usb_uninitialize;
    dev->base_dev.fnc_ioctl=g65ddc_usb_ioctl;
    dev->base_dev.fnc_open_locked=g65ddc_usb_open_locked;
    dev->base_dev.fnc_close_locked=g65ddc_usb_close_locked;
    dev->base_dev.fnc_set_power_locked=g65ddc_usb_set_power_locked;
    dev->base_dev.fnc_check_temperature_locked=g65ddc_usb_check_temperature_locked;

#ifdef G65DDC_COMPAT
    dev->base_dev.fnc_ioctl_compat=g65ddc_usb_ioctl_compat;
#endif

    dev->led_mode=USB_LED_MODE_DIAG;

    dev->base_dev.vid_pid[0]=G65DDC_USB_VENDOR_ID;
    dev->base_dev.vid_pid[1]=G65DDC_USB_PRODUCT_ID;

    init_rwsem(&dev->rw_sem);

    for(i=0;i<dev->ep_count;i++)
    {
        init_usb_anchor(&dev->ep[i].anchor);
        INIT_LIST_HEAD(&dev->ep[i].buffer_list);
        spin_lock_init(&dev->ep[i].spinlock);
        init_waitqueue_head(&dev->ep[i].wait_queue);
    }

    return 0;
}

static int g65ddc_usb_setup(struct g65ddc_usb *dev)
{
    g65ddc_usb_update_led_pattern(dev);

    return 0;
}

static int g65ddc_usb_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
 int status;
 struct g65ddc_usb *dev;
 struct usb_device *usb_dev;
 uint8_t i,ep_count;
 uint8_t ep_address;
 struct usb_endpoint_descriptor *ep_desc;
 int j;

    usb_dev=interface_to_usbdev(interface);

    dev_print_info(&usb_dev->dev,"probe");

    if(unlikely(id->idVendor!=G65DDC_USB_VENDOR_ID || id->idProduct!=G65DDC_USB_PRODUCT_ID))
    {
        dev_print_error(&usb_dev->dev,"unsupported device!");
        return -ENODEV;
    }

    if(unlikely(usb_dev->speed<USB_SPEED_HIGH))
    {
        dev_print_error(&usb_dev->dev,"device is not running at supported speed!");
        return -ENODEV;
    }

    ep_count=interface->cur_altsetting->desc.bNumEndpoints;

    dev=kzalloc(sizeof(*dev)+sizeof(*dev->ep)*ep_count,GFP_KERNEL);

    if(unlikely(dev==NULL))
    {
        print_error("out of memory!");
        return -ENOMEM;
    }

    dev->ep=(struct g65ddc_usb_ep*)(dev+1);

    status=usb_control_msg(usb_dev,usb_rcvctrlpipe(usb_dev,0),0xCC,USB_TYPE_VENDOR|USB_DIR_IN,0,0,dev->base_dev.fw,3*sizeof(uint16_t),G65DDC_USB_TIMEOUT);

    if(unlikely(status<0))
    {
        kfree(dev);
        dev_print_error(&usb_dev->dev,"failed to read device information!");
        return -ENODEV;
    }

    status/=sizeof(uint16_t);

    for(j=0;j<status;j++)
    {
        dev->base_dev.fw[j]=le16_to_cpu(dev->base_dev.fw[j]);
    }

    if(unlikely(usb_dev->speed<USB_SPEED_SUPER))
    {
        dev_print_warn(&usb_dev->dev,"device is not running at its maximum speed!");
        dev->base_dev.bus_type=BUS_TYPE_USB2;
    }
    else
    {
        dev->base_dev.bus_type=BUS_TYPE_USB3;
    }

    dev->usb_dev=usb_get_dev(usb_dev);
    dev->base_dev.p_dev=&usb_dev->dev;
    dev->usb_interface=usb_get_intf(interface);
    dev->ep_count=ep_count;

    usb_set_intfdata(interface,dev);

    dev->snd_ctrl_pipe=usb_sndctrlpipe(usb_dev,0);
    dev->rcv_ctrl_pipe=usb_rcvctrlpipe(usb_dev,0);

    for(i=0;i<ep_count;i++)
    {
        ep_desc=&interface->cur_altsetting->endpoint[i].desc;

        ep_address=ep_desc->bEndpointAddress;

        dev->ep[i].dev=dev;
        dev->ep[i].address=ep_address;
        dev->ep[i].interval=ep_desc->bInterval;
        dev->ep[i].type=usb_endpoint_type(ep_desc);
        dev->ep[i].max_packet_size=usb_endpoint_maxp(ep_desc);

        switch(dev->ep[i].type)
        {
            case USB_ENDPOINT_XFER_CONTROL:
                dev->ep[i].pipe=(ep_address&USB_DIR_IN)?usb_rcvctrlpipe(usb_dev,ep_address):usb_sndctrlpipe(usb_dev,ep_address);
                break;

            case USB_ENDPOINT_XFER_ISOC:
                dev->ep[i].pipe=(ep_address&USB_DIR_IN)?usb_rcvisocpipe(usb_dev,ep_address):usb_sndisocpipe(usb_dev,ep_address);
                break;

            case USB_ENDPOINT_XFER_BULK:
                dev->ep[i].pipe=(ep_address&USB_DIR_IN)?usb_rcvbulkpipe(usb_dev,ep_address):usb_sndbulkpipe(usb_dev,ep_address);
                break;

            case USB_ENDPOINT_XFER_INT:
                dev->ep[i].pipe=(ep_address&USB_DIR_IN)?usb_rcvintpipe(usb_dev,ep_address):usb_sndintpipe(usb_dev,ep_address);
                break;
        }
    }

    status=g65ddc_usb_initialize(dev);

    if(unlikely(status!=0))
    {
        g65ddc_uninitialize(&dev->base_dev);
        usb_set_intfdata(interface,NULL);
        usb_put_intf(dev->usb_interface);
        usb_put_dev(dev->usb_dev);
        kfree(dev);
        return status;
    }

    status=g65ddc_register(&dev->base_dev);

    if(unlikely(status<0))
    {
        g65ddc_delete(&dev->base_dev);
        return status;
    }

    g65ddc_usb_setup(dev);

    return 0;
}

static void g65ddc_usb_disconnect(struct usb_interface *interface)
{
 struct g65ddc_usb *dev;

    dev=usb_get_intfdata(interface);

    dev_print_info(&dev->usb_dev->dev,"disconnect");

    usb_set_intfdata(interface,NULL);

    if(dev)
    {
        dev_usb_lock(dev);

        SET_STATE(&dev->base_dev,connected=0);

        g65ddc_usb_set_power_locked(&dev->base_dev,0);

        dev_usb_unlock(dev);

        g65ddc_cancel_wd_work(&dev->base_dev);

        g65ddc_unregister(&dev->base_dev);

        usb_lock_exclusive(dev);

        dev->usb_interface=NULL;

        usb_unlock_exclusive(dev);
    }
}


static int g65ddc_usb_suspend(struct usb_interface *interface,pm_message_t message)
{
 struct g65ddc_usb *dev;

    dev=usb_get_intfdata(interface);

    if(likely(dev!=NULL))
    {
        dev_usb_lock(dev);

        if(dev->base_dev.state.s.power)
        {
            dev->base_dev.state.s.suspended=1;
            g65ddc_usb_set_power_locked(&dev->base_dev,0);
        }

        dev_usb_unlock(dev);
    }

    return 0;
}

static int g65ddc_usb_resume(struct usb_interface *interface)
{
 struct g65ddc_usb *dev;

    dev=usb_get_intfdata(interface);

    if(likely(dev!=NULL))
    {
        g65ddc_usb_setup(dev);
    }

    return 0;
}



static struct usb_device_id g65ddc_usb_id_table[]=
{
    {USB_DEVICE(G65DDC_USB_VENDOR_ID,G65DDC_USB_PRODUCT_ID)},
    {}
};

MODULE_DEVICE_TABLE(usb,g65ddc_usb_id_table);

static struct usb_driver g65ddc_usb_driver=
{
    .name=G65DDC_MODULE_NAME,
    .probe=g65ddc_usb_probe,
    .disconnect=g65ddc_usb_disconnect,
    .id_table=g65ddc_usb_id_table,
    .suspend=g65ddc_usb_suspend,
    .resume=g65ddc_usb_resume,
    .reset_resume=g65ddc_usb_resume
};

#undef USB_DEV

#endif //G65DDC_USB
#ifdef G65DDC_PCI

#define irq_spin_lock(dev,flags) spin_lock_irqsave(&dev->irq_spinlock,flags)
#define irq_spin_unlock(dev,flags) spin_unlock_irqrestore(&dev->irq_spinlock,flags)

#define to_g65ddc_pci(__base_dev__) container_of(__base_dev__,struct g65ddc_pci,base_dev)

static inline void dev_pci_lock(struct g65ddc_pci *dev)
{
    dev_lock(&dev->base_dev);
}

static inline void dev_pci_unlock(struct g65ddc_pci *dev)
{
    dev_unlock(&dev->base_dev);
}

static inline void pci_lock_shared(struct g65ddc_pci *dev)
{
    down_read(&dev->rw_sem);
}

static inline void pci_unlock_shared(struct g65ddc_pci *dev)
{
    up_read(&dev->rw_sem);
}

static inline void pci_lock_exclusive(struct g65ddc_pci *dev)
{
    down_write(&dev->rw_sem);
}

static inline void pci_unlock_exclusive(struct g65ddc_pci *dev)
{
    up_write(&dev->rw_sem);
}

static inline void dma_lock(struct g65ddc_pci_dma_channel *dma_channel)
{
    mutex_lock(&dma_channel->mutex);
}

static inline void dma_unlock(struct g65ddc_pci_dma_channel *dma_channel)
{
    mutex_unlock(&dma_channel->mutex);
}

static void g65ddc_pci_dma_complete_buffer(struct g65ddc_pci *dev,struct g65ddc_pci_dma_channel *dma_channel,struct g65ddc_pci_dma_buffer *buffer,uint32_t bytes_trasferred)
{
    dma_sync_sg_for_cpu(&dev->pci_dev->dev,buffer->io_buffer.sg_list,buffer->io_buffer.page_count,dma_channel->direction);
    dma_unmap_sg(&dev->pci_dev->dev,buffer->io_buffer.sg_list,buffer->io_buffer.page_count,dma_channel->direction);

    if(dma_channel->direction==DMA_FROM_DEVICE)
    {
        buffer->io_buffer.changed=true;
    }

    buffer->dma_map_sg_count=0;
    buffer->io_buffer.status=0;
    buffer->io_buffer.bytes_transferred=bytes_trasferred;
    buffer->io_buffer.idle=true;

    if(list_first_entry_or_null(&dma_channel->buffer_list,typeof(*buffer),io_buffer.list_entry)==buffer)
    {
        wake_up(&dma_channel->wait_queue);
    }
}

static void g65ddc_pci_dma_start_transfer(struct g65ddc_pci *dev,struct g65ddc_pci_dma_channel *dma_channel,struct g65ddc_pci_dma_buffer *buffer)
{
 struct g65ddc_pci_dma_sg_element *element;
 struct g65ddc_pci_dma_ptr *ptr;
 uint32_t ptr_index;
 struct scatterlist *sg;
 uint32_t i;

    ptr_index=dma_channel->ptr_idle_index;

    dma_channel->ptr_idle_index=NEXT_PTR_INDEX(dma_channel->ptr_idle_index);

    dma_channel->ptr_active_count++;

    ptr=dma_channel->ptr+ptr_index;

    element=ptr->sg_addr_cpu;

    for_each_sg(buffer->io_buffer.sg_list,sg,buffer->dma_map_sg_count,i)
    {
        element[i].address=sg_dma_address(sg);
        element[i].size=sg_dma_len(sg);
    }

    ptr->buffer=buffer;

    dma_sync_sg_for_device(&dev->pci_dev->dev,buffer->io_buffer.sg_list,buffer->io_buffer.page_count,dma_channel->direction);

    bar2_write32(ptr->reg+0,(uint32_t)(ptr->sg_addr_dma & 0xFFFFFFFF));
    bar2_write32(ptr->reg+4,(uint32_t)(ptr->sg_addr_dma>>32));
    bar2_write32(ptr->reg+8,buffer->dma_map_sg_count);
}

static void g65ddc_pci_dma_finish_transfer(struct g65ddc_pci *dev,struct g65ddc_pci_dma_channel *dma_channel,uint32_t int_stat)
{
 uint32_t i,count;
 struct g65ddc_pci_dma_buffer *buffer;

    dma_lock(dma_channel);

    if(dma_channel->active)
    {
        count=dma_channel->ptr_active_count;

        for(i=0;i<count;i++)
        {
            if(int_stat & (1<<dma_channel->ptr_active_index))
            {
                if(dma_channel->ptr[dma_channel->ptr_active_index].buffer==NULL)
                {
                    break;
                }

                g65ddc_pci_dma_complete_buffer(dev,dma_channel,dma_channel->ptr[dma_channel->ptr_active_index].buffer,dma_channel->buffer_size);

                dma_channel->ptr[dma_channel->ptr_active_index].buffer=NULL;
                dma_channel->ptr_active_count--;
                dma_channel->ptr_active_index=NEXT_PTR_INDEX(dma_channel->ptr_active_index);

                if(likely(!list_empty(&dma_channel->buffer_list_pending)))
                {
                    buffer=list_first_entry(&dma_channel->buffer_list_pending,typeof(*buffer),list_entry_pending);

                    list_del(&buffer->list_entry_pending);

                    g65ddc_pci_dma_start_transfer(dev,dma_channel,buffer);
                }
            }
        }
    }

    dma_unlock(dma_channel);
}

static void g65ddc_pci_irq_work_function(struct work_struct *work)
{
 struct g65ddc_pci *dev;
 unsigned long flags;
 uint32_t int_stat_from_device;
 uint32_t int_stat_to_device;
 uint32_t i;

    dev=container_of(work,struct g65ddc_pci,irq_work);

    irq_spin_lock(dev,flags);

    int_stat_from_device=dev->dma_from_device.int_stat & dev->dma_from_device.int_mask;
    int_stat_to_device=dev->dma_to_device.int_stat & dev->dma_to_device.int_mask;

    dev->dma_from_device.int_stat=0;
    dev->dma_to_device.int_stat=0;

    irq_spin_unlock(dev,flags);

    for(i=0;int_stat_from_device!=0 && i<PCI_DMA_CHANNEL_COUNT_FROM_DEVICE;i++,int_stat_from_device>>=2)
    {
        if(int_stat_from_device&3)
        {
            g65ddc_pci_dma_finish_transfer(dev,dev->dma_from_device.channel+i,int_stat_from_device);
        }
    }

    for(i=0;int_stat_to_device!=0 && i<PCI_DMA_CHANNEL_COUNT_TO_DEVICE;i++,int_stat_to_device>>=2)
    {
        if(int_stat_to_device&3)
        {
            g65ddc_pci_dma_finish_transfer(dev,dev->dma_to_device.channel+i,int_stat_to_device);
        }
    }
}


static irqreturn_t g65ddc_pci_interrupt_handler(int irq_no,void *dev_id)
{
 struct g65ddc_pci *dev;
 uint32_t int_stat;
 unsigned long flags;

    if(unlikely(dev_id==NULL))
    {
        return IRQ_NONE;
    }

    dev=dev_id;

    irq_spin_lock(dev,flags);

    if(unlikely(dev->base_dev.state.s.suspended))
    {
        irq_spin_unlock(dev,flags);
        return IRQ_NONE;
    }

    int_stat=bar2_read32(REG_PCI_INT_STAT) & dev->int_mask;

    if(unlikely(int_stat==0))
    {
        irq_spin_unlock(dev,flags);
        return IRQ_NONE;
    }

    if(int_stat  & (INT_DMA_OUT|INT_DMA_IN))
    {
        if(int_stat & INT_DMA_OUT)
        {
            dev->dma_from_device.int_stat|=bar2_read32(REG_PCI_DMA_INT_STAT_OUT) & dev->dma_from_device.int_mask;
        }

        if(int_stat & INT_DMA_IN)
        {
            dev->dma_to_device.int_stat|=bar2_read32(REG_PCI_DMA_INT_STAT_IN) & dev->dma_to_device.int_mask;
        }

        queue_work(dev->base_dev.workqueue,&dev->irq_work);
    }

    bar2_write32(REG_PCI_INT_STAT,int_stat);
    bar2_read32(REG_PCI_INT_STAT);

    irq_spin_unlock(dev,flags);

    return IRQ_HANDLED;
}

static void g65ddc_pci_dma_enable_channel(struct g65ddc_pci *dev,struct g65ddc_pci_dma_channel *dma_channel)
{
 unsigned long flags;

    irq_spin_lock(dev,flags);

    if(dma_channel->direction==DMA_FROM_DEVICE)
    {
        dev->dma_from_device.enabled|=1<<dma_channel->index;
        bar2_write32(REG_PCI_DMA_ENABLE_OUT,dev->dma_from_device.enabled);
        dev->dma_from_device.int_mask|=3<<(dma_channel->index<<1);
        bar2_read32(REG_PCI_DMA_ENABLE_OUT);

        if((dev->int_mask & INT_DMA_OUT)==0)
        {
            dev->int_mask|=INT_DMA_OUT;
            bar2_write32(REG_PCI_INT_MASK,dev->int_mask);
            bar2_read32(REG_PCI_INT_MASK);
        }
    }
    else
    {
        dev->dma_to_device.enabled|=1<<dma_channel->index;
        bar2_write32(REG_PCI_DMA_ENABLE_IN,dev->dma_to_device.enabled);
        dev->dma_to_device.int_mask|=3<<(dma_channel->index<<1);
        bar2_read32(REG_PCI_DMA_ENABLE_IN);

        if((dev->int_mask & INT_DMA_IN)==0)
        {
            dev->int_mask|=INT_DMA_IN;
            bar2_write32(REG_PCI_INT_MASK,dev->int_mask);
            bar2_read32(REG_PCI_INT_MASK);
        }
    }

    irq_spin_unlock(dev,flags);
}

static void g65ddc_pci_dma_disable_channel(struct g65ddc_pci *dev,struct g65ddc_pci_dma_channel *dma_channel)
{
 unsigned long flags;

    irq_spin_lock(dev,flags);

    if(dma_channel->direction==DMA_FROM_DEVICE)
    {
        dev->dma_from_device.enabled&=~(1<<dma_channel->index);
        bar2_write32(REG_PCI_DMA_ENABLE_OUT,dev->dma_from_device.enabled);
        dev->dma_from_device.int_mask&=~(3<<(dma_channel->index<<1));
        dev->dma_from_device.int_stat&=dev->dma_from_device.int_mask;
        bar2_read32(REG_PCI_DMA_ENABLE_OUT);

        if(dev->dma_from_device.int_mask==0 && (dev->int_mask & INT_DMA_OUT)!=0)
        {
            dev->int_mask&=~INT_DMA_OUT;
            bar2_write32(REG_PCI_INT_MASK,dev->int_mask);
            bar2_read32(REG_PCI_INT_MASK);
        }
    }
    else
    {
        dev->dma_to_device.enabled&=~(1<<dma_channel->index);
        bar2_write32(REG_PCI_DMA_ENABLE_IN,dev->dma_to_device.enabled);
        dev->dma_to_device.int_mask&=~(3<<(dma_channel->index<<1));
        dev->dma_to_device.int_stat&=dev->dma_to_device.int_mask;
        bar2_read32(REG_PCI_DMA_ENABLE_IN);

        if(dev->dma_to_device.int_mask==0 && (dev->int_mask & INT_DMA_IN)!=0)
        {
            dev->int_mask&=~INT_DMA_IN;
            bar2_write32(REG_PCI_INT_MASK,dev->int_mask);
            bar2_read32(REG_PCI_INT_MASK);
        }
    }

    irq_spin_unlock(dev,flags);
}

static struct g65ddc_pci_dma_channel* g65ddc_pci_get_dma_channel(struct g65ddc_pci *dev,uint16_t direction,uint16_t channel_index)
{
    switch(direction)
    {
        case DMA_DIRECTION_FROM_DEVICE:
            if(unlikely(channel_index>=ARRAY_SIZE(dev->dma_from_device.channel)))
            {
                return NULL;
            }
            return dev->dma_from_device.channel+channel_index;

        case DMA_DIRECTION_TO_DEVICE:
            if(unlikely(channel_index>=ARRAY_SIZE(dev->dma_to_device.channel)))
            {
                return NULL;
            }
            return dev->dma_to_device.channel+channel_index;

        default:
            return NULL;
    }
}

static int g65ddc_pci_dma_start(struct g65ddc_pci *dev,uint16_t direction,uint16_t channel_index,uint32_t buffer_size)
{
 struct g65ddc_pci_dma_channel *dma_channel;
 uint32_t sg_mem_size;
 uint32_t i;

    if(buffer_size==0 || (buffer_size&PCI_BUFFER_SIZE_ALIGNMENT_MASK)!=0)
    {
        return -EINVAL;
    }

    dma_channel=g65ddc_pci_get_dma_channel(dev,direction,channel_index);

    if(unlikely(dma_channel==NULL))
    {
        return -EINVAL;
    }

    sg_mem_size=DIV_ROUND_UP(buffer_size,PAGE_SIZE)*sizeof(struct g65ddc_pci_dma_sg_element);

    dev_pci_lock(dev);

    if(dev->base_dev.state.s.power==0)
    {
        dev_pci_unlock(dev);
        return -EPERM;
    }

    for(i=0;i<DMA_CHANNEL_PTR_COUNT;i++)
    {
        dma_channel->ptr[i].sg_addr_cpu=dma_alloc_coherent(&dev->pci_dev->dev,sg_mem_size,&dma_channel->ptr[i].sg_addr_dma,GFP_ATOMIC);

        if(unlikely(dma_channel->ptr[i].sg_addr_cpu==NULL))
        {
            while(i--)
            {
                dma_free_coherent(&dev->pci_dev->dev,sg_mem_size,dma_channel->ptr[i].sg_addr_cpu,dma_channel->ptr[i].sg_addr_dma);
                dma_channel->ptr[i].sg_addr_cpu=NULL;
                dma_channel->ptr[i].sg_addr_dma=0;
            }

            dev_pci_unlock(dev);

            return -ENOMEM;
        }
    }

    dma_channel->buffer_size=buffer_size;
    dma_channel->sg_mem_size=sg_mem_size;

    g65ddc_pci_dma_enable_channel(dev,dma_channel);

    dma_channel->active=true;

    dev_pci_unlock(dev);

    return 0;
}


static void g65ddc_pci_dma_stop(struct g65ddc_pci *dev,struct g65ddc_pci_dma_channel *dma_channel)
{
 struct g65ddc_pci_dma_buffer *buffer;
 int i;

    dma_lock(dma_channel);

    if(dma_channel->active)
    {
        g65ddc_pci_dma_disable_channel(dev,dma_channel);

        dma_channel->active=false;

        for(i=0;i<DMA_CHANNEL_PTR_COUNT;i++)
        {
            if(dma_channel->ptr[i].buffer)
            {
                g65ddc_pci_dma_complete_buffer(dev,dma_channel,dma_channel->ptr[i].buffer,0);
                dma_channel->ptr[i].buffer=NULL;
            }
        }

        dma_channel->ptr_active_count=0;
        dma_channel->ptr_active_index=0;
        dma_channel->ptr_idle_index=0;

        list_for_each_entry(buffer,&dma_channel->buffer_list_pending,list_entry_pending)
        {
            g65ddc_pci_dma_complete_buffer(dev,dma_channel,buffer,0);
        }

        INIT_LIST_HEAD(&dma_channel->buffer_list_pending);

        list_for_each_entry(buffer,&dma_channel->buffer_list,io_buffer.list_entry)
        {
            g65ddc_io_buffer_unprepare(&buffer->io_buffer);
            kfree(buffer);
        }

        INIT_LIST_HEAD(&dma_channel->buffer_list);

        for(i=0;i<DMA_CHANNEL_PTR_COUNT;i++)
        {
            dma_free_coherent(&dev->pci_dev->dev,dma_channel->sg_mem_size,dma_channel->ptr[i].sg_addr_cpu,dma_channel->ptr[i].sg_addr_dma);
            dma_channel->ptr[i].sg_addr_cpu=NULL;
            dma_channel->ptr[i].sg_addr_dma=0;
        }

        wake_up(&dma_channel->wait_queue);
    }

    dma_unlock(dma_channel);
}

static int g65ddc_pci_dma_get_complete_buffer(struct g65ddc_pci *dev,uint16_t direction,uint16_t channel_index,uintptr_t *user_buffer)
{
 struct g65ddc_pci_dma_channel *dma_channel;
 struct g65ddc_pci_dma_buffer *buffer;
 uint32_t bytes_transferred;
 DEFINE_WAIT(wait);

    dma_channel=g65ddc_pci_get_dma_channel(dev,direction,channel_index);

    if(unlikely(dma_channel==NULL))
    {
        return -EINVAL;
    }

    for(;;)
    {
        dma_lock(dma_channel);

        if(unlikely(!dma_channel->active))
        {
            dma_unlock(dma_channel);
            return -EPERM;
        }

        if(!list_empty(&dma_channel->buffer_list))
        {
            buffer=list_first_entry(&dma_channel->buffer_list,typeof(*buffer),io_buffer.list_entry);

            if(buffer->io_buffer.idle)
            {
                list_del(&buffer->io_buffer.list_entry);
                dma_unlock(dma_channel);

                *user_buffer=buffer->io_buffer.user_buffer;

                bytes_transferred=buffer->io_buffer.bytes_transferred;

                g65ddc_io_buffer_unprepare(&buffer->io_buffer);
                kfree(buffer);

                return bytes_transferred;
            }
        }

        prepare_to_wait(&dma_channel->wait_queue,&wait,TASK_INTERRUPTIBLE);

        dma_unlock(dma_channel);

        schedule();

        finish_wait(&dma_channel->wait_queue,&wait);

        if(unlikely(signal_pending(current)))
        {
            return -ERESTARTSYS;
        }
    }
}

static int g65ddc_pci_ioctl_dma_get_complete_buffer(struct g65ddc_pci *dev,void __user *user_ptr)
{
 union g65ddc_ioctl_pci_dma_complete_buffer tmp;
 uintptr_t user_buffer;
 int status;

    if(unlikely(copy_from_user(&tmp.channel,user_ptr,sizeof(tmp.channel))))
    {
        return -EFAULT;
    }

    status=g65ddc_pci_dma_get_complete_buffer(dev,tmp.channel.direction,tmp.channel.channel_index,&user_buffer);

    if(likely(status>=0))
    {
        if(unlikely(copy_to_user((void __user*)user_ptr,&user_buffer,sizeof(tmp.buffer))))
        {
            return -EFAULT;
        }
    }

    return status;
}

static int g65ddc_pci_dma_queue_buffer(struct g65ddc_pci *dev,uint16_t direction,uint16_t channel_index,uintptr_t user_buffer,uint32_t size)
{
 struct g65ddc_pci_dma_channel *dma_channel;
 struct g65ddc_pci_dma_buffer *buffer;
 int status;

    dma_channel=g65ddc_pci_get_dma_channel(dev,direction,channel_index);

    if(unlikely(dma_channel==NULL))
    {
        return -EINVAL;
    }

    if(unlikely(!check_user_buffer(user_buffer,size)))
    {
        return -EINVAL;
    }

    dma_lock(dma_channel);

    if(unlikely(!dma_channel->active))
    {
        dma_unlock(dma_channel);
        return -EPERM;
    }

    if(unlikely(dma_channel->buffer_size!=size))
    {
        dma_unlock(dma_channel);
        return -EINVAL;
    }

    buffer=kmalloc(sizeof(*buffer),GFP_KERNEL);

    if(unlikely(buffer==NULL))
    {
        dma_unlock(dma_channel);
        return -ENOMEM;
    }

    status=g65ddc_io_buffer_prepare(&buffer->io_buffer,user_buffer,size,direction==DMA_DIRECTION_FROM_DEVICE);

    if(unlikely(status!=0))
    {
        dma_unlock(dma_channel);
        kfree(buffer);
        return status;
    }

    buffer->dma_map_sg_count=dma_map_sg(&dev->pci_dev->dev,buffer->io_buffer.sg_list,buffer->io_buffer.page_count,dma_channel->direction);

    if(unlikely(buffer->dma_map_sg_count==0))
    {
        dma_unlock(dma_channel);
        g65ddc_io_buffer_unprepare(&buffer->io_buffer);
        kfree(buffer);
        return -EIO;
    }

    list_add_tail(&buffer->io_buffer.list_entry,&dma_channel->buffer_list);

    buffer->io_buffer.idle=false;
    buffer->io_buffer.bytes_transferred=0;
    buffer->io_buffer.status=0;

    dma_sync_sg_for_device(&dev->pci_dev->dev,buffer->io_buffer.sg_list,buffer->io_buffer.page_count,dma_channel->direction);

    if(dma_channel->ptr_active_count<DMA_CHANNEL_PTR_COUNT)
    {
        g65ddc_pci_dma_start_transfer(dev,dma_channel,buffer);
    }
    else
    {
        list_add_tail(&buffer->list_entry_pending,&dma_channel->buffer_list_pending);
    }

    dma_unlock(dma_channel);

    return 0;
}


static int g65ddc_pci_ioctl_dma_queue_buffer(struct g65ddc_pci *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_pci_dma_queue_buffer queue_buffer_params;

    if(unlikely(copy_from_user(&queue_buffer_params,user_ptr,sizeof(queue_buffer_params))))
    {
        return -EFAULT;
    }

    return g65ddc_pci_dma_queue_buffer(dev,queue_buffer_params.direction,queue_buffer_params.channel_index,(uintptr_t)queue_buffer_params.buffer,queue_buffer_params.size);
}

static int g65ddc_pci_ioctl_dma_start(struct g65ddc_pci *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_pci_dma_start_params params;

    if(unlikely(copy_from_user(&params,user_ptr,sizeof(params))))
    {
        return -EINVAL;
    }

    return g65ddc_pci_dma_start(dev,params.direction,params.channel_index,params.buffer_size);
}

static int g65ddc_pci_ioctl_dma_stop(struct g65ddc_pci *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_pci_dma_channel dma_channel_params;
 struct g65ddc_pci_dma_channel *dma_channel;

    if(unlikely(copy_from_user(&dma_channel_params,user_ptr,sizeof(dma_channel_params))))
    {
        return -EFAULT;
    }

    dma_channel=g65ddc_pci_get_dma_channel(dev,dma_channel_params.direction,dma_channel_params.channel_index);

    if(unlikely(dma_channel==NULL))
    {
        return -EINVAL;
    }

    g65ddc_pci_dma_stop(dev,dma_channel);

    return 0;
}

static long g65ddc_pci_ioctl(struct g65ddc *base_dev,struct file *file,unsigned int cmd,unsigned long arg)
{
    switch(cmd)
    {
        case G65DDC_IOCTL_PCI_DMA_GET_COMPLETE_BUFFER:
            return g65ddc_pci_ioctl_dma_get_complete_buffer(to_g65ddc_pci(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_PCI_DMA_QUEUE_BUFFER:
            return g65ddc_pci_ioctl_dma_queue_buffer(to_g65ddc_pci(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_PCI_DMA_START:
            return g65ddc_pci_ioctl_dma_start(to_g65ddc_pci(base_dev),(void __user*)arg);

        case G65DDC_IOCTL_PCI_DMA_STOP:
            return g65ddc_pci_ioctl_dma_stop(to_g65ddc_pci(base_dev),(void __user*)arg);

        default:
            return g65ddc_ioctl(base_dev,file,cmd,arg);
    }
}

#ifdef G65DDC_COMPAT

static int g65ddc_pci_ioctl_dma_get_complete_buffer_compat(struct g65ddc_pci *dev,void __user *user_ptr)
{
 union g65ddc_ioctl_pci_dma_complete_buffer_compat tmp;
 uintptr_t user_buffer;
 int status;

    if(unlikely(copy_from_user(&tmp.channel,user_ptr,sizeof(tmp.channel))))
    {
        return -EFAULT;
    }

    status=g65ddc_pci_dma_get_complete_buffer(dev,tmp.channel.direction,tmp.channel.channel_index,&user_buffer);

    if(likely(status>=0))
    {
        if(unlikely(copy_to_user((void __user*)user_ptr,&user_buffer,sizeof(tmp.buffer))))
        {
            return -EFAULT;
        }
    }

    return status;
}

static int g65ddc_pci_ioctl_dma_queue_buffer_compat(struct g65ddc_pci *dev,void __user *user_ptr)
{
 struct g65ddc_ioctl_pci_dma_queue_buffer_compat queue_buffer_params;

    if(unlikely(copy_from_user(&queue_buffer_params,user_ptr,sizeof(queue_buffer_params))))
    {
        return -EFAULT;
    }

    return g65ddc_pci_dma_queue_buffer(dev,queue_buffer_params.direction,queue_buffer_params.channel_index,(uintptr_t)compat_ptr(queue_buffer_params.buffer),queue_buffer_params.size);
}

static long g65ddc_pci_ioctl_compat(struct g65ddc *base_dev,struct file *file, unsigned int cmd, unsigned long arg)
{
    switch(cmd)
    {
        case G65DDC_IOCTL_PCI_DMA_GET_COMPLETE_BUFFER_COMPAT:
            return g65ddc_pci_ioctl_dma_get_complete_buffer_compat(to_g65ddc_pci(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_PCI_DMA_QUEUE_BUFFER_COMPAT:
            return g65ddc_pci_ioctl_dma_queue_buffer_compat(to_g65ddc_pci(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_PCI_DMA_START:
            return g65ddc_pci_ioctl_dma_start(to_g65ddc_pci(base_dev),compat_ptr(arg));

        case G65DDC_IOCTL_PCI_DMA_STOP:
            return g65ddc_pci_ioctl_dma_stop(to_g65ddc_pci(base_dev),compat_ptr(arg));

        default:
            return g65ddc_ioctl_compat(base_dev,file,cmd,arg);
    }
}

#endif //G65DDC_COMPAT

static inline bool g65ddc_pci_check_temperature(struct g65ddc_pci *dev,uint32_t reg,uint32_t critical_temperature)
{
 int temperature;

    temperature=bar2_read32(reg);

    temperature*=2500;
    temperature-=500<<10;
    temperature+=5<<10;
    temperature/=10<<10;

    return temperature>=critical_temperature;
}


static int g65ddc_pci_check_temperature_locked(struct g65ddc *base_dev,bool *critical_temperature)
{
 struct g65ddc_pci *dev;

    dev=to_g65ddc_pci(base_dev);

    *critical_temperature=g65ddc_pci_check_temperature(dev,REG_PCI_TEMP_SENSOR_0,CRITICAL_TEMPERATURE_0) ||
        g65ddc_pci_check_temperature(dev,REG_PCI_TEMP_SENSOR_1,CRITICAL_TEMPERATURE_1) ||
        g65ddc_pci_check_temperature(dev,REG_PCI_TEMP_SENSOR_2,CRITICAL_TEMPERATURE_2);

    return 0;
}

static int g65ddc_pci_set_power_locked(struct g65ddc *base_dev,uint8_t power)
{
 size_t i;
 struct g65ddc_pci *dev;

    if(base_dev->state.s.power==power)
    {
        return 0;
    }

    dev=to_g65ddc_pci(base_dev);

    if(power==0)
    {
        SET_STATE(base_dev,power=0);

        for(i=0;i<ARRAY_SIZE(dev->dma_from_device.channel);i++)
        {
            g65ddc_pci_dma_stop(dev,dev->dma_from_device.channel+i);
        }

        for(i=0;i<ARRAY_SIZE(dev->dma_to_device.channel);i++)
        {
            g65ddc_pci_dma_stop(dev,dev->dma_to_device.channel+i);
        }

        bar2_write32(REG_PCI_FPGA_CFG_CTRL,0);
        bar2_read32(REG_PCI_FPGA_CFG_CTRL);

        dev->fpga=false;
    }

    bar2_write32(REG_PCI_HW_CTRL,(power>>1)&3);
    bar2_read32(REG_PCI_HW_CTRL);

    if(power)
    {
        SET_STATE(base_dev,power=power);
    }

    return 0;
}

static int g65ddc_pci_open_locked(struct g65ddc *base_dev)
{
    return 0;
}

static void g65ddc_pci_close_locked(struct g65ddc *base_dev)
{
 struct g65ddc_pci *dev;

    g65ddc_pci_set_power_locked(base_dev,0);

    dev=to_g65ddc_pci(base_dev);

    if((bar2_read32(0x0F08)&0x401)!=0x401)
    {
        bar2_write32(0x0F08,0x403);
    }
}

static void g65ddc_pci_uninitialize(struct g65ddc *base_dev)
{
}

static void g65ddc_pci_delete(struct g65ddc *base_dev)
{
 size_t i;
 struct g65ddc_pci *dev;

    dev=to_g65ddc_pci(base_dev);

    if(dev->irq_handler_registered)
    {
        free_irq(pci_irq_vector(dev->pci_dev,0),dev);
    }

    pci_free_irq_vectors(dev->pci_dev);

    pci_iounmap(dev->pci_dev,dev->bar2);

    pci_release_regions(dev->pci_dev);
    pci_disable_device(dev->pci_dev);

    for(i=0;i<ARRAY_SIZE(dev->dma_from_device.channel);i++)
    {
        mutex_destroy(&dev->dma_from_device.channel[i].mutex);
    }

    for(i=0;i<ARRAY_SIZE(dev->dma_to_device.channel);i++)
    {
        mutex_destroy(&dev->dma_to_device.channel[i].mutex);
    }

    pci_dev_put(dev->pci_dev);
}

static int g65ddc_pci_mmap(struct g65ddc *base_dev,struct vm_area_struct *vma)
{
 struct g65ddc_pci *dev;
 unsigned long size;
 uint32_t i;
 int status;
 int bars[]={0,2};

    dev=to_g65ddc_pci(base_dev);

    if(unlikely(vma->vm_pgoff!=0))
    {
        return -EINVAL;
    }

    size=vma->vm_end-vma->vm_start;

    for(i=0;i<ARRAY_SIZE(bars);i++)
    {
        if(size==pci_resource_len(dev->pci_dev,bars[i]))
        {
            status=io_remap_pfn_range(vma,vma->vm_start,pci_resource_start(dev->pci_dev,bars[i])>>PAGE_SHIFT,size,vma->vm_page_prot);

            if(unlikely(status!=0))
            {
                return -EAGAIN;
            }

            return 0;
        }
    }

    return -EINVAL;
}

static int g65ddc_pci_initialize(struct g65ddc_pci *dev)
{
 int status;
 size_t i;
 int j;

    status=g65ddc_initialize(&dev->base_dev);

    if(unlikely(status!=0))
    {
        return status;
    }

    dev->base_dev.fnc_delete=g65ddc_pci_delete;
    dev->base_dev.fnc_uninitialize=g65ddc_pci_uninitialize;
    dev->base_dev.fnc_ioctl=g65ddc_pci_ioctl;
    dev->base_dev.fnc_open_locked=g65ddc_pci_open_locked;
    dev->base_dev.fnc_close_locked=g65ddc_pci_close_locked;
    dev->base_dev.fnc_set_power_locked=g65ddc_pci_set_power_locked;
    dev->base_dev.fnc_check_temperature_locked=g65ddc_pci_check_temperature_locked;
    dev->base_dev.fnc_mmap=g65ddc_pci_mmap;

#ifdef G65DDC_COMPAT
    dev->base_dev.fnc_ioctl_compat=g65ddc_pci_ioctl_compat;
#endif

    dev->base_dev.vid_pid[0]=G65DDC_PCI_VENDOR_ID;
    dev->base_dev.vid_pid[1]=G65DDC_PCI_DEVICE_ID;

    spin_lock_init(&dev->irq_spinlock);

    INIT_WORK(&dev->irq_work,g65ddc_pci_irq_work_function);

    for(i=0;i<ARRAY_SIZE(dev->dma_from_device.channel);i++)
    {
        dev->dma_from_device.channel[i].index=i;
        dev->dma_from_device.channel[i].direction=DMA_FROM_DEVICE;
        INIT_LIST_HEAD(&dev->dma_from_device.channel[i].buffer_list);
        INIT_LIST_HEAD(&dev->dma_from_device.channel[i].buffer_list_pending);
        init_waitqueue_head(&dev->dma_from_device.channel[i].wait_queue);
        mutex_init(&dev->dma_from_device.channel[i].mutex);

        for(j=0;j<DMA_CHANNEL_PTR_COUNT;j++)
        {
            dev->dma_from_device.channel[i].ptr[j].reg=0x05A0+i*6*4+j*3*4;
        }
    }

    for(i=0;i<ARRAY_SIZE(dev->dma_to_device.channel);i++)
    {
        dev->dma_to_device.channel[i].index=i;
        dev->dma_to_device.channel[i].direction=DMA_TO_DEVICE;
        INIT_LIST_HEAD(&dev->dma_to_device.channel[i].buffer_list);
        INIT_LIST_HEAD(&dev->dma_to_device.channel[i].buffer_list_pending);
        init_waitqueue_head(&dev->dma_to_device.channel[i].wait_queue);
        mutex_init(&dev->dma_to_device.channel[i].mutex);

        for(j=0;j<DMA_CHANNEL_PTR_COUNT;j++)
        {
            dev->dma_to_device.channel[i].ptr[j].reg=0x0420+i*6*4+j*3*4;
        }
    }

    dev->base_dev.fw[0]=(uint16_t)bar2_read32(0x0088);
    dev->base_dev.fw[1]=(uint16_t)bar2_read32(0x0F14);

    return 0;
}


static int g65ddc_pci_probe(struct pci_dev *pci_dev,const struct pci_device_id *id)
{
 struct g65ddc_pci *dev;
 int status;

    dev_print_info(&pci_dev->dev,"probe");

    if(unlikely(pci_dev->vendor!=G65DDC_PCI_VENDOR_ID || pci_dev->device!=G65DDC_PCI_DEVICE_ID))
    {
        dev_print_error(&pci_dev->dev,"unsupported device!");
        return -ENODEV;
    }

    dev=kzalloc(sizeof(*dev),GFP_KERNEL);

    if(unlikely(dev==NULL))
    {
        dev_print_error(&pci_dev->dev,"out of memory!");
        return -ENOMEM;
    }

    dev->pci_dev=pci_dev_get(pci_dev);
    dev->base_dev.p_dev=&pci_dev->dev;
    dev->base_dev.bus_type=BUS_TYPE_PCIE;


    status=pci_enable_device(dev->pci_dev);

    if(unlikely(status!=0))
    {
        dev_print_error(&pci_dev->dev,"failed to enable device (%d)!",status);
        goto error_kfree;
    }

    pci_set_drvdata(dev->pci_dev,dev);

    if(likely(dma_set_mask_and_coherent(&dev->pci_dev->dev,DMA_BIT_MASK(64))!=0))
    {
        if(unlikely(dma_set_mask_and_coherent(&dev->pci_dev->dev,DMA_BIT_MASK(32))!=0))
        {
            dev_print_error(&pci_dev->dev,"no suitable DMA available!");
            goto error_pci_disable_device;
        }
    }

    pci_set_master(dev->pci_dev);

    status=pci_request_regions(dev->pci_dev,pci_name(dev->pci_dev));

    if(unlikely(status!=0))
    {
        dev_print_error(&pci_dev->dev,"failed to reserve PCI I/O memory resources (%d)!",status);
        goto error_pci_disable_device;
    }

    dev->bar2=pci_iomap(dev->pci_dev,2,0);

    if(unlikely(dev->bar2==NULL))
    {
        dev_print_error(&pci_dev->dev,"failed to map resource!");
        status=-EIO;
        goto error_pci_release_regions;
    }

    status=pci_alloc_irq_vectors(dev->pci_dev,1,1,PCI_IRQ_LEGACY|PCI_IRQ_MSI);

    if(unlikely(status<1))
    {
        dev_print_error(&pci_dev->dev,"failed to allocate irq vector (%d)!",status);
        status=-EIO;
        goto error_unmap_bar2;
    }

    status=g65ddc_pci_initialize(dev);

    if(unlikely(status!=0))
    {
        goto error_free_irq_vectors;
    }

    bar2_write32(REG_PCI_INT_MASK,0);
    bar2_write32(REG_PCI_DMA_ENABLE_IN,0);
    bar2_write32(REG_PCI_DMA_ENABLE_OUT,0);

    status=request_irq(pci_irq_vector(pci_dev,0),g65ddc_pci_interrupt_handler,IRQF_SHARED,G65DDC_MODULE_NAME,dev);

    if(unlikely(status<0))
    {
        dev_print_error(&pci_dev->dev,"failed to register interrupt handler (%d)!",status);
        g65ddc_delete(&dev->base_dev);
        return status;
    }

    dev->irq_handler_registered=true;

    status=g65ddc_register(&dev->base_dev);

    if(unlikely(status<0))
    {
        g65ddc_delete(&dev->base_dev);
        return status;
    }

    return 0;

    error_free_irq_vectors:
        pci_free_irq_vectors(dev->pci_dev);

    error_unmap_bar2:
        pci_iounmap(dev->pci_dev,dev->bar2);

    error_pci_release_regions:
        pci_release_regions(dev->pci_dev);

    error_pci_disable_device:
        pci_set_drvdata(dev->pci_dev,NULL);
        pci_disable_device(dev->pci_dev);

    error_kfree:
        pci_dev_put(dev->pci_dev);
        kfree(dev);

    return status;
}


static void g65ddc_pci_remove(struct pci_dev *pci_dev)
{
 struct g65ddc_pci *dev;

    dev=pci_get_drvdata(pci_dev);

    if(likely(dev!=NULL))
    {
        SET_STATE(&dev->base_dev,connected=0);

        g65ddc_cancel_wd_work(&dev->base_dev);

        pci_set_drvdata(pci_dev,NULL);

        g65ddc_unregister(&dev->base_dev);
    }
}

static int g65ddc_pci_suspend(struct pci_dev *pci_dev,pm_message_t state)
{
 struct g65ddc_pci *dev;

    dev=pci_get_drvdata(pci_dev);

    if(likely(dev!=NULL))
    {
        dev_pci_lock(dev);

        dev->base_dev.state.s.suspended=1;
        g65ddc_pci_set_power_locked(&dev->base_dev,0);


        pci_save_state(pci_dev);
        pci_disable_device(pci_dev);
        pci_set_power_state(pci_dev,pci_choose_state(pci_dev,state));

        dev_pci_unlock(dev);
    }

    return 0;
}

static int g65ddc_pci_resume(struct pci_dev *pci_dev)
{
 struct g65ddc_pci *dev;
 int status;

    dev=pci_get_drvdata(pci_dev);

    if(likely(dev!=NULL))
    {
        pci_set_power_state(pci_dev,PCI_D0);
        pci_restore_state(pci_dev);

        status=pci_enable_device(pci_dev);

        if(unlikely(status!=0))
        {
            dev_print_error(&pci_dev->dev,"failed to enable device for resume (%d)!",status);
            return status;
        }

        pci_set_master(pci_dev);

        dev_pci_lock(dev);

        dev->base_dev.state.s.suspended=0;

        dev_pci_unlock(dev);
    }

    return 0;
}

static struct pci_device_id g65ddc_pci_id_table[]=
{
    {PCI_DEVICE(G65DDC_PCI_VENDOR_ID,G65DDC_PCI_DEVICE_ID)},
    {}
};

static struct pci_driver g65ddc_pci_driver=
{
    .name=G65DDC_MODULE_NAME,
    .probe=g65ddc_pci_probe,
    .remove=g65ddc_pci_remove,
    .id_table=g65ddc_pci_id_table,
    .suspend=g65ddc_pci_suspend,
    .resume=g65ddc_pci_resume,
};

MODULE_DEVICE_TABLE(pci,g65ddc_pci_id_table);

#endif //G65DDC_PCI


static int __init g65ddc_init(void)
{
 int status;

    status=alloc_chrdev_region(&base_device_number,0,G65DDC_MAX_DEVICE_COUNT,G65DDC_MODULE_NAME);

    if(status!=0)
    {
        print_error("could not register a range of device numbers");
        return status;
    }

    memset(registered_devices,0,sizeof(registered_devices));
    mutex_init(&reg_mutex);

    status=class_register(&g65ddc_class);

    if(status<0)
    {
        print_error("failed to register class!");
        goto error_unregister_chrdev_region;
    }

#ifdef G65DDC_USB

    status=usb_register(&g65ddc_usb_driver);

    if(status!=0)
    {
        print_error("failed to register usb driver!");
        goto error_class_unregister;
    }

#endif //G65DDC_USB

#ifdef G65DDC_PCI
    status=pci_register_driver(&g65ddc_pci_driver);

    if(status!=0)
    {
        print_error("failed to register pci driver!");
#ifdef G65DDC_USB
        usb_deregister(&g65ddc_usb_driver);
#endif //G65DDC_USB
        goto error_class_unregister;
    }

#endif //G65DDC_PCI

    return 0;

    error_class_unregister:
        class_unregister(&g65ddc_class);

    error_unregister_chrdev_region:
        unregister_chrdev_region(base_device_number,G65DDC_MAX_DEVICE_COUNT);

    return status;
}

static void __exit g65ddc_exit(void)
{
#ifdef G65DDC_PCI
    pci_unregister_driver(&g65ddc_pci_driver);
#endif //G65DDC_PCI

#ifdef G65DDC_USB
    usb_deregister(&g65ddc_usb_driver);
#endif //G65DDC_USB

    class_unregister(&g65ddc_class);
    unregister_chrdev_region(base_device_number,G65DDC_MAX_DEVICE_COUNT);

    mutex_destroy(&reg_mutex);
}

module_init(g65ddc_init);
module_exit(g65ddc_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR(G65DDC_AUTHOR);
MODULE_DESCRIPTION(G65DDC_DESCRIPTION);
